#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2018-12-26
#
# 2019-08-03: changes for cmk 1.5:  inv_tree_list to inv_tree, comment out split for cmk 1.5 
#                    #bug['known_fixed_releases'] = bug['known_fixed_releases'].split(' ')
#                    #bug['known_affected_releases'] = bug['known_affected_releases'].split(' ')
# 2021-07-24: rewritten for CMK 2.0
# 2021-07-26: some cleanup
#

import os
import json
import logging
import time
from typing import List, Optional
from dataclasses import dataclass
from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    SNMPTree,
    contains,
    any_of,
    OIDEnd,
    startswith,
    TableRow,
    Attributes,
)

from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    StringTable,
    InventoryResult,
)

from cmk.base.plugins.agent_based.utils.inv_cisco_support import (
    set_loglevel,
    expand_path,
    get_base_path,
)


@dataclass
class SnmpBug:
    version: str
    pid: str


def _get_os_version(cw_version, sysdescription, inventory):
    pid = ''
    sys_version = None
    phy_version = None

    if str(cw_version).find('CW_VERSION$') == 0:
        cw_version = cw_version.split('$')[1]  # .upper()

    sysdescription = sysdescription.split(',')
    # sys description from N5K-C5672UP (has two version entries :-( )
    # Cisco NX-OS(tm) n6000, Software (n6000-uk9), Version 7.0(8)N1(1), RELEASE SOFTWARE Copyright (c) 2002-2012 \
    # by Cisco Systems, Inc. Device Manager Version 6.0(2)N1(1), Compiled 2/20/2016 21:00:00
    for entry in sysdescription:
        if not sys_version and 'version' in entry.lower():  # get only the first 'version' entry
            sys_version = entry[entry.lower().find('version') + 7:].strip()
            if sys_version.startswith(':'):  # AsyncOS
                sys_version = sys_version[1:].strip()
            sys_version = sys_version.split(' ')[0].strip()
            sys_version = sys_version  # .upper()

    low_index = 1000000
    for phyentry in inventory:
        phyindex, physoftwarerev, phymodelname, physicalclass = phyentry
        if physicalclass == '3':  # chassis
            if int(phyindex) <= low_index:
                if phymodelname != '':  # PID at all
                    if phymodelname.find(' ') == -1:  # no spaces in PID
                        low_index = int(phyindex)
                        pid = phymodelname
                        phy_version = physoftwarerev.split(',')[0]

    version = cw_version
    if version == '' and sys_version:
        version = sys_version
    if version == '' and phy_version:
        version = phy_version

    # remove leading '0' in IOSXE version numbers
    if version and version[0] == '0':
        version = version.replace('.0', '.')
        version = version.lstrip('0')
    version = version.rstrip(';')

    # todo: cut intermediate for ASA version as option

    return version, pid


def _get_bug_status(bug_status):
    _bug_status = {
        'F': 'Fixed',
        'O': 'Open',
        'T': 'Terminated'
    }
    return _bug_status.get(bug_status, bug_status)


def _get_bug_behavior_changed(behavior_changed):
    _behavior_changed = {
        'Y': 'yes',
        'N': 'no',
    }
    return _behavior_changed.get(behavior_changed, behavior_changed)


def parse_inv_cisco_bug(string_table: List[StringTable]) -> Optional[SnmpBug]:
    cw_version, sysdescription = string_table[0][0]
    pids = string_table[1]

    # get OS version
    version, pid = _get_os_version(cw_version, sysdescription, pids)

    if version == '' or pid == '':
        return
    else:
        return SnmpBug(
            version=version,
            pid=pid
        )


def inventory_cisco_bug(params, section: SnmpBug) -> InventoryResult:
    set_loglevel()

    optionalcolumns = [
        'base_pid',
        'description',
        'id',
        'known_affected_releases',
        'product',
    ]

    # get parameters from wato
    if params:
        optionalcolumns = params.get('removecolumns', optionalcolumns)

    base_path = get_base_path()
    bug_dir = base_path + '/bug'
    path_found = expand_path(bug_dir + '/found/')
    path_not_found = expand_path(bug_dir + '/not_found/')
    path_request = expand_path(bug_dir + '/request/')

    path = ['software', 'support', 'cisco_bug']
    attribute = {
        'os_version': section.version,
        'PID': section.pid,
    }

    pid = section.pid.replace('/', '_')

    # bug report found
    if os.path.isfile(path_found + pid + '/' + section.version):
        modifytime = os.path.getmtime(path_found + pid + '/' + section.version)
        attribute.update({
            'Last_checked': time.strftime('%Y-%m-%d', time.localtime(modifytime)),
            'status': 'found'
        })

        bug_record = ''
        with open(path_found + pid + '/' + section.version) as f:
            try:
                bug_record = json.load(f)
            except ValueError as e:
                logging.warning(f'inv_cisco_bug:bug_found:JSON load error: {e}')
                exit()
        if bug_record != '':
            attribute['Total_records'] = bug_record.get('total_records', 0)

            missing = bug_record.get('missing', {})
            if len(missing.keys()) != 0:
                attribute['missing_records'] = len(missing.keys()) * 10

            bugs = bug_record.get('bugs')
            bug_ids = []
            clean_bugs = []
            for bug in bugs:
                if bug.get('bug_id') not in bug_ids:
                    bug_ids.append(bug.get('bug_id'))
                    bug['known_fixed_releases'] = ', '.join(bug['known_fixed_releases'].split(' '))
                    bug['known_affected_releases'] = ', '.join(bug['known_affected_releases'].split(' '))
                    bug['status'] = _get_bug_status(bug['status'])
                    bug['behavior_changed'] = _get_bug_behavior_changed(bug['behavior_changed'])
                    if optionalcolumns is not None:
                        for column in optionalcolumns:
                            bug.pop(column, None)
                    clean_bugs.append(bug)
                    key_columns = {'bug_id': bug.get('bug_id')}
                    bug.pop('bug_id')
                    yield TableRow(
                        path=path + ['bugs'],
                        key_columns=key_columns,
                        inventory_columns=bug,
                    )
            attribute['duplicate_records'] = len(bugs) - len(clean_bugs)

    # bug report not found
    elif os.path.isfile(path_not_found + pid + '/' + section.version):
        modifytime = os.path.getmtime(path_not_found + pid + '/' + section.version)
        attribute.update({
            'Last_checked': time.strftime('%Y-%m-%d', time.localtime(modifytime)),
            'status': 'Not found',
        })
    else:
        # create new request
        path_request = expand_path(path_request + pid)
        if not os.path.isfile(path_request + '/' + section.version):
            with open(path_request + '/' + section.version, 'w+') as f:  # only create file (touch)
                pass
        modifytime = os.path.getmtime(path_request + '/' + section.version)
        attribute.update({
            'Last_checked': time.strftime('%Y-%m-%d', time.localtime(modifytime)),
            'status': 'requested',
        })

    yield Attributes(
        path=path,
        inventory_attributes=attribute
    )


register.snmp_section(
    name='inv_cisco_bug',
    parse_function=parse_inv_cisco_bug,
    fetch=[
        SNMPTree(
            base='.1.3.6.1',  #
            oids=[
                '4.1.9.9.25.1.1.1.2.5',  # CISCO-IMAGE-MIB::CW_VERSION
                '2.1.1.1.0',  # sysDescription
            ]),
        SNMPTree(
            base='.1.3.6.1.2.1.47.1.1.1.1',  # ENTITY-MIB::entPhysicalEntry
            oids=[
                OIDEnd(),  # index of entry
                '10',  # entPhysicalSoftwareRev
                '13',  # entPhysicalModelName
                '5',  # entPhysicalClass
            ]),
    ],

    detect=any_of(
        contains('.1.3.6.1.2.1.1.1.0', 'cisco'),
        startswith('.1.3.6.1.2.1.1.2.0', '.1.3.6.1.4.1.9.1')
    )
)

register.inventory_plugin(
    name='inv_cisco_bug',
    inventory_function=inventory_cisco_bug,
    inventory_default_parameters={},
    inventory_ruleset_name='inv_cisco_bug',
)
