#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2017-07-05
#
# collect/show Cisco psirt information for IOS/IOSXE
#
# 2018-04-09: removed import
# 2018-09-04: changes for CMK 1.5.x (inv_tree --> inv_tree_list)
# 2018-09-05: changes for CMK 1.5.x (replaced global variable g_hostname with api call host_name())
# 2018-01-06: minor changes, added status info for removed advisories (not updated, older than, ...)
# 2019-08-13: changed for cmk 1.5
#             #node = inv_tree_list('software.support.cisco_psirt.')
#              node = inv_tree('software.support.cisco_psirt.')
#             changed variable host_name to _hostname for cmk 1.5
# 2021-07-24: rewritten for CMK 2.0
# 2021-07-26: cleanup
#
# ToDo: add number of psirt records to status information
#

import os
import json
import time
from typing import List
from dataclasses import dataclass

from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    SNMPTree,
    contains,
    any_of,
    startswith,
    TableRow,
    Attributes,
)

from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    StringTable,
    InventoryResult,
)

from cmk.base.plugins.agent_based.utils.inv_cisco_support import (
    expand_path,
    get_base_path,
)


@dataclass
class SnmpPsirt:
    version: str
    product_family: str


def _create_psirt_record(filepath, filename, not_updated, dont_show_older_then, optionalcolumns):
    psirtfile = filepath + filename
    advisories = {}
    if os.path.isfile(psirtfile):
        with open(psirtfile) as f:
            try:
                psirtrecord = json.load(f)
            except ValueError:
                exit()

        advisories = psirtrecord.get('advisories')
        remove_advisories = []
        for advisory in advisories:
            last_updated = advisory.get('lastUpdated', 'T1980-01-01').split('T')[0]
            if last_updated == '':
                last_updated = '1980-01-01'
            if (not_updated > int(
                    (time.time() - time.mktime(time.strptime(last_updated, '%Y-%m-%d'))) / 86400)) and (
                    last_updated >= dont_show_older_then):
                advisory.update({'installed_version': filename})
                advisory.update({'lastUpdated': advisory.get('lastUpdated', 'T1980-01-01').split('T')[0]})
                advisory.update({'firstPublished': advisory.get('firstPublished', 'T1980-01-01').split('T')[0]})
                advisory.update({'bugIDs': ', '.join(advisory.get('bugIDs', ''))})
                advisory.update({'firstFixed': ', '.join(advisory.get('firstFixed', ''))})
                advisory.update({'cves': ', '.join(advisory.get('cves', ''))})
                advisory.update({'cwe': ', '.join(advisory.get('cwe', ''))})

                keys = advisory.keys()
                for column in optionalcolumns:
                    if column in keys:
                        advisory.pop(column)
            else:
                # add old advisories to remove list
                remove_advisories.insert(0, advisories.index(advisory))

        # remove advisories older then not_updated
        # if len(remove_advisories) > 0:
        #    node['removed_advisories'] = len(remove_advisories)
        # for index in remove_advisories:
        #    advisories.pop(index)
    return advisories


def _get_profuct_family(sysdescription: str) -> str:
    if ' IOS ' in sysdescription:
        if ' IOS-XE ' in sysdescription or 'IOSXE' in sysdescription or 'IOS Software [' in sysdescription:
            return 'IOS-XE'
        else:
            return 'IOS'
    elif 'NX-OS' in sysdescription:
        if ' aci' in sysdescription or 'APIC ' in sysdescription:
            return 'APIC'
        else:
            return 'NEXUS'
    elif 'APIC ' in sysdescription:
        return 'APIC'
    elif 'Cisco Controller' in sysdescription:
        return 'WLC'
    elif 'Cisco Adaptive Security Appliance Version' in sysdescription:
        return 'ASA'
    elif 'Cisco PIX Security Appliance Version' in sysdescription:
        return 'PIX'
    elif ' AsyncOS ' in sysdescription:
        if 'Model C' in sysdescription:
            return 'ESA'
        elif 'Model M' in sysdescription:
            return 'SMA'
        elif 'Model S' in sysdescription:
            return 'WSA'
    elif 'Integrated Management Controller' in sysdescription:
        return 'IMC'
    else:
        return 'no_family_found_or_set'


def _get_os_version(cw_version, sysdescription, pids):
    version = None

    if str(cw_version).find('CW_VERSION$') == 0:
        version = cw_version.split('$')[1]  # .upper()
    else:
        sysdescription = sysdescription.split(',')
        for entry in sysdescription:
            # if entry.strip().lower().startswith('version'):
            if 'version' in entry.lower():
                version = entry[entry.lower().find('version') + 7:].strip()
                if version.startswith(':'):  # AsyncOS
                    version = version[1:].strip()
                version = version.split(' ')[0].strip()
                version = version  # .upper()
        if not version:
            for pid in pids:
                physoftwarerev, phymodelname, physicalclass = pid
                if physicalclass == '3':  # chassies
                    version = physoftwarerev
    # remove leading '0' in IOSXE version numbers
    if version and version[0] == '0':
        version = version.replace('.0', '.')
        version = version.lstrip('0')

    return version


def parse_inv_cisco_psirt(string_table: List[StringTable]) -> SnmpPsirt:

    cw_version, sysdescription = string_table[0][0]

    pids = string_table[1]

    return SnmpPsirt(
        version=_get_os_version(cw_version, sysdescription, pids),
        product_family=_get_profuct_family(sysdescription)
    )


def inventory_cisco_psirt(params, section: SnmpPsirt) -> InventoryResult:
    optionalcolumns = [
        'publicationUrl',
        'summary',
        'cwe',
        'cves',
        'productNames',
        'iosRelease',
        'ipsSignatures',
        'cvrfUrl',
        'ovalUrl',
    ]
    product_family = None
    psirt_type = None
    not_updated = 10000000
    dont_show_older_then = '2000-01-01'

    # get parameters from wato
    if params:
        optionalcolumns = params.get('removecolumns', optionalcolumns)
        psirt_type = params.get('psirt_type', psirt_type)
        if type(psirt_type) == dict:
            product_family = psirt_type.get('product_family')  # .upper()
            not_updated = psirt_type.get('not_updated', not_updated)
            dont_show_older_then = psirt_type.get('dont_show_older_then')

    if not product_family:
        # try to get product family form system description
        product_family = section.product_family
    elif psirt_type in ['IOS', 'IOS-XE']:
        product_family = psirt_type

    # set paths
    base_path = get_base_path()
    psirt_dir = base_path + '/psirt'
    path_found = psirt_dir + '/found/'
    path_not_found = psirt_dir + '/not_found/'
    path_request = psirt_dir + '/request/'

    if product_family in ['IOS', 'IOS-XE']:
        path_found = path_found + product_family
        path_not_found = path_not_found + product_family
        path_request = path_request + product_family
        psirt = section.version
    else:
        path_found = path_found + 'family'
        path_not_found = path_not_found + 'family'
        path_request = path_request + 'family'
        psirt = product_family

    path_found = expand_path(path_found)
    path_not_found = expand_path(path_not_found)
    path_request = expand_path(path_request)

    # set defaults for inventory entries
    path = ['software', 'support', 'cisco_psirt']
    attribute = {
        'os_version': section.version,
        'Product_family': product_family,
    }

    # psirt record found
    if os.path.isfile(path_found + psirt):
        modifytime = os.path.getmtime(path_found + psirt)
        attribute.update({
            'status': 'found',
            'Last_checked': time.strftime('%Y-%m-%d', time.localtime(modifytime)),
        })

        # create psirt advisory list
        if product_family in ['IOS', 'IOS-XE']:  # do not remove entry's from IOS(XE)
            advisories = _create_psirt_record(path_found, psirt, 10000000, '2000-01-01', optionalcolumns)
        else:
            advisories = _create_psirt_record(path_found, psirt, not_updated, dont_show_older_then, optionalcolumns)
            if dont_show_older_then != '0000-00-00':
                attribute['dont_show_older_than'] = dont_show_older_then
            if not_updated != 10000000:
                attribute['dont_show_not_updated_since'] = not_updated

        for advisory in advisories:
            key_columns = {'advisoryId': advisory.get('advisoryId')}
            advisory.pop('advisoryId')
            yield TableRow(
                path=path + ['advisories'],
                key_columns=key_columns,
                inventory_columns=advisory,
            )

    # psirt record not found exists
    elif os.path.isfile(path_not_found + psirt):
        modifytime = os.path.getmtime(path_not_found + psirt)
        attribute.update({
            'status': 'not found',
            'Last_checked': time.strftime('%Y-%m-%d', time.localtime(modifytime)),
        })
    elif psirt != 'no_family_found_or_set':
        # create new request
        if not os.path.isfile(path_request + psirt):
            with open(path_request + psirt, 'w+') as f:  # only create file (touch)
                pass
        modifytime = os.path.getmtime(path_request + psirt)
        attribute.update({
            'status': 'requested',
            'Last_checked': time.strftime('%Y-%m-%d', time.localtime(modifytime)),
        })

    yield Attributes(
        path=path,
        inventory_attributes=attribute,
    )


register.snmp_section(
    name='inv_cisco_psirt',
    parse_function=parse_inv_cisco_psirt,
    fetch=[
        SNMPTree(
            base='.1.3.6.1',  #
            oids=[
                '4.1.9.9.25.1.1.1.2.5',  # CISCO-IMAGE-MIB::CW_VERSION
                '2.1.1.1.0',  # sysDescription
            ]),
        SNMPTree(
            base='.1.3.6.1.2.1.47.1.1.1.1',  # ENTITY-MIB::entPhysicalEntry
            oids=[
                '10',  # entPhysicalSoftwareRev
                '13',  # entPhysicalModelName
                '5',  # entPhysicalClass
            ]),
    ],

    detect=any_of(
        contains('.1.3.6.1.2.1.1.1.0', 'cisco'),
        startswith('.1.3.6.1.2.1.1.2.0', '.1.3.6.1.4.1.9.1')
    )
)

register.inventory_plugin(
    name='inv_cisco_psirt',
    inventory_function=inventory_cisco_psirt,
    inventory_default_parameters={},
    inventory_ruleset_name='inv_cisco_psirt',
)
