#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2017-03-20
#
# include file, will be used with snmp_cisco_eox and snmp_cisco_contract
#
# 2017-05-29: fixed empty pid handling
#             added serial number cleanup
# 2021-07-23: rewrite for CMK 2.0
# 2021-07-30: renamed form ciscoapi.py to inv_cisco_support.py

import os
import logging
import re
import json
from typing import List, Dict
from dataclasses import dataclass

from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    StringTable,
)


@dataclass
class SnmpContractEntry:
    phydescr: str
    pid: str


# list of PIDs to drop
g_PID_black_list = ['BUILT-IN', 'MICRON', 'C400-MTFDD']
# list of PIDs to try by serial number
g_PID_bad_list = ['UNSPECIFIED', 'FABRIC', 'ASA', 'C2611XM-2FE', 'FTLX8570D3BCL', 'FTLF8519P2BCL', 'FTLX8571D3BCL',
                  'FTRJ-8519-7D', 'PLRXPL-SC-S43']  #
# list of S/Ns to drop
g_SN_black_list = []


def parse_inv_cisco_contract_eox(string_table: StringTable) -> Dict[str, SnmpContractEntry]:
    serials: Dict[str, SnmpContractEntry] = {}
    for phydescr, physerialnum, phymodelname in string_table:
        if not check_bad_serial(physerialnum):
            if not pid_on_black_list(phymodelname):
                phymodelname = phymodelname.strip()
                physerialnum = physerialnum.strip()
                if phymodelname == '':
                    phymodelname = 'UNSPECIFIED'
                pid = phymodelname.split()[0].upper()
                serial = physerialnum.split()[0].upper()
                if serial not in serials.keys():
                    serials.update({serial: SnmpContractEntry(
                        phydescr=phydescr,
                        pid=pid
                    )})
                else:
                    temp = serials[serial]
                    temp.phydescr = ', '.join(list(set(temp.phydescr.split(', ') + [phydescr])))
                    temp.pid = ', '.join(list(set(temp.pid.split(', )') + [pid])))
                    serials[serial] = temp

    return serials


def get_base_path() -> str:
    conf_file = os.path.expanduser('~/etc/ciscoapi/ciscoapi.json')
    base_path = '~/var/ciscoapi'
    # check for conf_file and read parameters
    if os.path.isfile(conf_file):
        with open(conf_file) as f:
            try:
                config = json.load(f)
                base_path = config['global'].get('base_dir', base_path)
            except ValueError as e:
                logging.warning(f'inv_cisco_contract:status:JSON load error: {e}')
    return base_path


def set_pid_black_list(pid_black_list: List[str]):
    global g_PID_black_list
    if pid_black_list:
        g_PID_black_list = list(set(g_PID_black_list + pid_black_list))


def set_pid_bad_list(pid_bad_list: List[str]):
    global g_PID_bad_list
    if pid_bad_list:
        g_PID_bad_list = list(set(g_PID_bad_list + pid_bad_list))


def set_sn_black_list(sn_black_list: List[str]):
    global g_SN_black_list
    if sn_black_list:
        g_SN_black_list = list(set(g_SN_black_list + sn_black_list))


def check_dir_and_create(directory: str) -> bool:
    """
    check if dir exists, if not try to create it.

    Args:
        directory:

    Returns:
        True: if dir exists or creation was ok
        False: if dir not exists and creation was not ok
    """
    directory = os.path.dirname(directory)
    if not os.path.exists(directory):
        try:
            os.makedirs(directory)
        except:
            return False
    return True


def expand_path(path: str) -> str:
    """
    expand user dir and add '/' if necessary and create directory if it not exists
    """
    homedir = os.path.expanduser('~')
    if path.startswith('~'):
        path = homedir + path[1:]
    if not path.endswith('/'):
        path += '/'
    if not check_dir_and_create(path):
        return ''
    return path


def sn_on_black_list(serial: str) -> bool:
    """
    Checks if serial on black list

    Args:
        serial: the serial number to check

    Returns:
        True: if serial number on black list
        False: if serial number is not on black list
    """
    global g_SN_black_list
    if serial.upper() in g_SN_black_list:
        return True
    return False


def pid_on_black_list(pid: str) -> bool:
    """
    returns True if pid on black list
    """
    global g_PID_black_list
    for drop_PID in g_PID_black_list:
        if pid.startswith(drop_PID.upper()):
            return True
    # if PID not on Black list return false
    return False


def pid_on_bad_list(pid: str) -> bool:
    """
    returns True if PID on Bad list
    """
    global g_PID_bad_list
    # remove all chars from string, except allowedchars
    allowedchars = re.compile('[^A-Z0-9_=\/\-\+\.\\\]')
    cleanpid = allowedchars.sub('', pid).strip().upper()

    # if PID contains illegal signs or if pid empty try by serial number
    if (cleanpid != pid) or (cleanpid == ''):
        return True

    # list of bad PIDs :-( we need to try get EoX info via serial number for this PIDs
    # needs to be configurable via a file if too big for wato ;-(
    for bad_PID in g_PID_bad_list:
        if pid.startswith(bad_PID):
            return True

    return False


def check_bad_serial(serial: str) -> bool:
    """
    Checks a serial number for illegal signs (anything except A-Z0-9_=-+.\)

    Args:
        serial: serial number to check

    Returns:
        False: if serial number is ok
        True: if serial number contains bad signs
    """
    logging.info('Check_bad_serial:serial: %s' % serial)
    serial = serial.replace(' ', '').upper()

    # remove all chars from string, except allowedchars
    allowedchars = re.compile('[^A-Z0-9_=\-\+\.\\\]')
    cleanserial = allowedchars.sub('', serial).strip()

    logging.info('Check_bad_serial:cleanserial: %s' % cleanserial)

    # if serial contains illegal signs or empty return true
    if (cleanserial != serial) or (cleanserial == ''):
        logging.info('Check_bad_serial:bad:serial is bad')

        return True

    logging.info('Check_bad_serial:serial is god')
    return False


def set_loglevel():
    # set default logglevel
    logging.basicConfig(level=logging.WARNING)
    # has no effect as long previous command is active (is by design)
    logging.basicConfig(level=logging.INFO)
