#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2017-07-10
#
#  https://developer.cisco.com/docs/support-apis/
#
# 2018-06-06: fixed handling if state changes form found to not_found (delete old psirt found file)
# 2021-07-24: rewritten for python 3.8
#
#
import ntpath
import os
import json
from typing import List
from dataclasses import dataclass

from cisco_live_cycle_utils import (
    configure_logger,
    log_message,
    get_ids_from_dir,
    remove_ids_from_list,
    refresh_ids_from_dir,
    sleep_random,
    expand_path,
)
from ciscoapi import (
    AccessToken,
    Settings,
    get_psirt_by_product_family,
    get_psirt_by_iosxe_version,
    get_psirt_by_ios_version,
)


@dataclass
class Paths:
    found: str
    not_found: str
    request: str


@dataclass
class Refresh:
    found: int
    not_found: int


g_logger = None


def _psirt_remove_id_file(psirt_path: str, psirt_id: str):
    # delete psirt file
    try:
        log_message(f'delete psirt id file : {psirt_path + psirt_id}')
        os.remove(psirt_path + psirt_id)
    except OSError:
        pass


def _psirt_dump_record(psirt_record, psirt_id: str, psirt_path: str, psirt_path_request: str):
    with open(psirt_path + psirt_id, 'w') as f:
        json.dump(psirt_record, f)
    # delete request file
    _psirt_remove_id_file(psirt_path_request, psirt_id)

    return


def _check_psirt_record(psirt_record, psirt_id, psirt_path_found, psirt_path_request):
    """

    :param psirt_record:
    :param psirt_id:
    :param psirt_path_found:
    :param psirt_path_request:
    :returns:
    """

    for advisory in psirt_record.get('advisories'):
        # remove unwanted information from advisories
        advisory.pop('productNames', None)
        advisory.pop('ipsSignatures', None)
        advisory.pop('iosRelease', None)
        advisory.pop('cvrfUrl', None)
        advisory.pop('ovalUrl', None)
        advisory.pop('summary', None)
        temp_advisory = advisory.copy()
        for key in temp_advisory.keys():
            if advisory.get(key, None) in [['NA'], 'NA']:
                advisory.pop(key, None)

    _psirt_dump_record(psirt_record, psirt_id, psirt_path_found, psirt_path_request)

    return


def _get_psirt_id_list(product_family: str, paths: Paths, refresh: Refresh) -> List[str]:
    """

    @param product_family:
    @param paths: Path object with path to founnd/not found/requested PSIRT records
    @param refresh: Refresh object with number of days before a PSIRT record needs to be refreshed for found/not found
    @return: list of PIDs
    """
    # create list of ID's to request PSIRT status for
    psirt_id_list = get_ids_from_dir(paths.request + product_family)
    log_message(f'psirt requests : {psirt_id_list}')
    # remove already found ID's from list
    psirt_id_list = remove_ids_from_list(psirt_id_list, paths.found + product_family)
    log_message(f'psirt requests : {psirt_id_list}')
    # remove not found ID's from list
    psirt_id_list = remove_ids_from_list(psirt_id_list, paths.not_found + product_family)
    log_message(f'psirt requests : {psirt_id_list}')

    # refresh psirt after 1 day by default
    psirt_id_list = refresh_ids_from_dir(paths.not_found + product_family, refresh.not_found, psirt_id_list, True)
    log_message(f'psirt requests : {psirt_id_list}')
    psirt_id_list = refresh_ids_from_dir(paths.found + product_family, refresh.found, psirt_id_list, False)
    log_message(f'psirt requests : {psirt_id_list}')

    return psirt_id_list


def _update_psirt_id(psirt_records: list, family_name: str, paths: Paths):
    for psirt_record in psirt_records:
        if family_name in ['IOS', 'IOS-XE']:
            psirt_id = psirt_record.get('version')
        else:
            psirt_id = psirt_record.get('family')

        if psirt_record.get('advisories') != 'notfound':
            _check_psirt_record(psirt_record, psirt_id, paths.found + family_name + '/',
                                ntpath.sep + family_name + '/')
        else:
            _psirt_dump_record(psirt_record, psirt_id, paths.not_found + family_name + '/',
                               paths.request + family_name + '/')
            # remove psirt_found file (happens when product family is removed form bug ID)
            _psirt_remove_id_file(paths.found + family_name + '/', psirt_id)
    return


def main():
    settings = Settings()
    access_token = AccessToken(settings.client_id, settings.client_secret)
    configure_logger(log_level=settings.log_level)

    refresh = Refresh(
        found=settings.psirt_refresh_found,
        not_found=settings.psirt_refresh_not_found
    )

    psirt_dir = expand_path(settings.base_path + '/psirt')
    paths = Paths(
        found=psirt_dir + '/found/',
        not_found=psirt_dir + '/not_found/',
        request=psirt_dir + '/request/'
    )

    psirt_ios = _get_psirt_id_list('IOS', paths, refresh)
    psirt_ios_xe = _get_psirt_id_list('IOS-XE', paths, refresh)
    psirt_family = _get_psirt_id_list('family', paths, refresh)

    if (psirt_ios == []) and psirt_ios_xe == [] and psirt_family == []:
        log_message('all list are empty. Do nothing.')
        return

    # wait random time after startup
    if settings.wait_after_start:
        sleep_random(settings.max_wait_time)

    if psirt_family is not []:
        psirt_records = get_psirt_by_product_family(psirt_family, access_token, settings=settings)
        _update_psirt_id(psirt_records, 'family', paths)

    if psirt_ios_xe is not []:
        psirt_records = get_psirt_by_iosxe_version(psirt_ios_xe, access_token, settings=settings)
        _update_psirt_id(psirt_records, 'IOS-XE', paths)

    if psirt_ios is not []:
        psirt_records = get_psirt_by_ios_version(psirt_ios, access_token, settings=settings)
        _update_psirt_id(psirt_records, 'IOS', paths)


main()
