#!/usr/bin/env python
# -*- encoding: utf-8; py-indent-offset: 4 -*-

#  https://developer.cisco.com/docs/support-apis/
#
# 2017-03-30: Cisco EoX API Framework
#             added support for Cisco SN2INFO API (get contract status)
# 2017-07-09: added support for Cisco Psirt API (IOS and IOSXE)
# 2017-07-19: added support for Cisco Software Suggestion API
# 2018-01-25: adding support for Cisco bug api 2.0
# 2018-09-25: performance improvement "psirt_response.encoding = 'UTF-8'", drops json.loads
#             from about 4 min to some seconds
# 2021-07-23: rewritten for python 3.8
# 2023-06-09: changed for new rest api endpoint (apix.cisco.com)
#             refactoring get_token and settings
#             some cleanup
#
# supportapis[dash]help[at]cisco[dot]com
#
import requests
import json
import time
from os.path import (
    expanduser,
)
from typing import Dict, List
from cisco_live_cycle_utils import (
    log_message,
)


class Settings:
    def __init__(self):
        conf_file = '~/etc/ciscoapi/ciscoapi.json'
        conf_file = expanduser(conf_file)

        with open(conf_file) as f:
            try:
                self.__settings = json.load(f)
            except ValueError as e:
                log_message(f'snmp_cisco_eox:status:JSON load error: {e}', level='WARNING')
                exit()
            except FileNotFoundError as e:
                log_message(f'Config file not found {e}.', level='CRITICAL')
                exit()

        self.__base_path = '~/var/ciscoapi'
        self.__auth_proxy_url = 'https://cmk.bech-noc.de/api/cauthproxy.py'
        self.__proxies = {}
        self.__wait_after_start = True
        self.__max_wait_time = 15
        self.__log_level = 'warning'
        self.__eox_refresh_known = 31
        self.__eox_refresh_unknown = 7
        self.__sn2info_refresh_covered = 31
        self.__sn2info_refresh_not_covered = 7
        self.__bug_refresh_found = 2
        self.__bug_refresh_not_found = 1
        self.__psirt_refresh_found = 1
        self.__psirt_refresh_not_found = 1
        self.__suggestion_refresh_found = 31
        self.__suggestion_refresh_not_found = 7

        if self.__settings['global'].get('http_proxy'):
            self.__proxies .update({'http': self.__settings['global'].get('http_proxy')})
        if self.__settings['global'].get('https_proxy'):
            self.__proxies .update({'https': self.__settings['global'].get('https_proxy')})

    @property
    def client_id(self) -> str:
        return self.__settings['cisco_api']['client_id']

    @property
    def client_secret(self) -> str:
        return self.__settings['cisco_api']['client_secret']

    @property
    def proxies(self) -> Dict[str, str]:
        return self.__proxies

    @property
    def use_system_proxies(self) -> bool:
        return self.__settings['global'].get('use_system_proxies', False)

    @property
    def use_auth_proxy(self) -> bool:
        return self.__settings['cisco_api'].get('use_auth_proxy', False)

    @property
    def client_fqdn(self) -> str:
        return self.__settings['cisco_api'].get('client_fqdn')

    @property
    def root_cert(self) -> bool:
        return self.__settings['cisco_api'].get('root_cert', False)

    @property
    def auth_proxy_url(self) -> str:
        return self.__settings['cisco_api'].get('auth_proxy_url')

    @property
    def base_path(self) -> str:
        return self.__settings['global'].get('base_path', self.__base_path)

    @property
    def wait_after_start(self) -> bool:
        return self.__settings['global'].get('wait_after_start', self.__wait_after_start)

    @property
    def max_wait_time(self) -> int:
        return self.__settings['global'].get('max_wait_time', self.__max_wait_time)

    @property
    def log_level(self) -> str:
        return self.__settings['global'].get('log_level', self.__log_level)

    @property
    def eox_refresh_known(self) -> int:
        return self.__settings['eox'].get('refresh_known', self.__eox_refresh_known)

    @property
    def eox_refresh_unknown(self) -> int:
        return self.__settings['eox'].get('refresh_known', self.__eox_refresh_unknown)

    @property
    def sn2info_refresh_covered(self) -> int:
        return self.__settings['sn2info'].get('refresh_covered', self.__sn2info_refresh_covered)

    @property
    def sn2info_refresh_not_covered(self) -> int:
        return self.__settings['sn2info'].get('refresh_not_covered', self.__sn2info_refresh_not_covered)

    @property
    def bug_refresh_found(self) -> int:
        return self.__settings['bug'].get('refresh_found', self.__bug_refresh_found)

    @property
    def bug_refresh_not_found(self) -> int:
        return self.__settings['bug'].get('refresh_found', self.__bug_refresh_not_found)

    @property
    def psirt_refresh_found(self) -> int:
        return self.__settings['psirt'].get('refresh_found', self.__psirt_refresh_found)

    @property
    def psirt_refresh_not_found(self) -> int:
        return self.__settings['psirt'].get('refresh_not_found', self.__psirt_refresh_not_found)

    @property
    def suggestion_refresh_found(self) -> int:
        return self.__settings['suggestion'].get('refresh_found', self.__suggestion_refresh_found)

    @property
    def suggestion_refresh_not_found(self) -> int:
        return self.__settings['suggestion'].get('refresh_not_found', self.__suggestion_refresh_not_found)


class AccessToken:
    def __init__(
            self,
            client_id: str,
            client_secret: str,
    ):
        self.__client_id = client_id
        self.__client_secret = client_secret
        self.__proxies = {}
        self.__use_auth_proxy = None
        self.__client_fqdn = ''
        self.__root_cert = True
        self.__auth_proxy_url = 'https://cmk.bech-noc.de/api/cauthproxy.py'
        self.__access_token = ''
        self.__lifetime = 0
        self.__time = 0
        self.__auth_headers = {
            'content-Type': 'application/x-www-form-urlencoded',
            'accept': 'application/json'
        }
        self.__grant_type = 'client_credentials'
        self.__auth_url = 'https://id.cisco.com/oauth2/default/v1/token'
        self.__verify = True
        self.__auth_req_data = {
            'client_id': self.__client_id,
            'client_secret': self.__client_secret,
            'grant_type': self.__grant_type
        }

    @property
    def token(self) -> str:

        if self.__access_token and time.time() < self.__lifetime:
            return self.__access_token
        else:
            self.__time = time.time()
            response = requests.post(
                self.__auth_url,
                headers=self.__auth_headers,
                data=self.__auth_req_data,
                proxies=self.__proxies,
                verify=self.__verify)

        if response.ok:
            auth_response = json.loads(response.text)
            self.__lifetime = self.__time + int(auth_response.get("expires_in")) - 60
            self.__access_token = auth_response.get("access_token")
            return self.__access_token


# generic cisco api request for all get info by serialnumber
def get_info_by_serials(
        serials: List[str],
        access_token: AccessToken,
        req_url: str,
        max_serials: int,
        settings: Settings
):
    # locale variablen
    max_serial_length = 40
    max_req_per_second = 5
    wait_time = 5
    optimisedserials = []
    serialsstr = ''
    count = 1
    info = []

    # split list of Serials in chunks of max 75 serials, each max 40 bytes length
    for serial in serials:
        if len(serial) <= max_serial_length:
            serialsstr += serial + ","
            count += 1
        if count == max_serials:
            optimisedserials.append(serialsstr[:-1])
            serialsstr = ''
            count = 1
    optimisedserials.append(serialsstr[:-1])

    headers = {'accept': 'application/json', 'Authorization': 'Bearer ' + access_token.token}

    count = 0
    for serials in optimisedserials:
        # Disable invalid certificate warnings.
        # requests.packages.urllib3.disable_warnings()
        response = requests.get(req_url + serials, headers=headers, proxies=settings.proxies)
        count += 1
        # only 5 request per second are allowed
        if count == max_req_per_second:
            time.sleep(wait_time)
            count = 0
        if response.ok:
            response.encoding = 'UTF-8'
            info.append(json.loads(response.text))

    return info


def get_eox_by_pid(pids: List[str], access_token: AccessToken, settings: Settings):
    # local variables
    max_pid_length = 240
    max_pids = 20
    max_req_per_second = 5
    wait_time = 5
    optimisedpids = []
    pidstr = ''
    count = 1
    eoxr = []

    # split list of PIDs in chunks of max 240 bytes length
    for pid in pids:
        if (len(pidstr) + len(pid)) < max_pid_length:
            pidstr += pid + ","
            count += 1
        if (count == max_pids) or ((len(pidstr) + len(pid)) >= max_pid_length):
            optimisedpids.append(pidstr[:-1])
            pidstr = ''
            count = 1
    optimisedpids.append(pidstr[:-1])
    headers = {'accept': 'application/json', 'Authorization': f'Bearer {access_token.token}'}
    req_url = 'https://apix.cisco.com/supporttools/eox/rest/5/EOXByProductID/1/'

    count = 0
    for productids in optimisedpids:
        # Disable invalid certificate warnings.
        # requests.packages.urllib3.disable_warnings()
        eoxresponse = requests.get(req_url + productids, headers=headers, proxies=settings.proxies)
        count += 1
        # only 5 request per second are allowed
        if count == max_req_per_second:
            time.sleep(wait_time)
            count = 0
        if eoxresponse.ok:
            eoxr.append(json.loads(eoxresponse.text))
    return eoxr


def get_eox_by_serials(serials: List[str], access_token: AccessToken, settings: Settings):
    max_serials = 20
    req_url = 'https://apix.cisco.com/supporttools/eox/rest/5/EOXBySerialNumber/1/'
    info = get_info_by_serials(serials, access_token, req_url, max_serials, settings)
    return info


def get_coverage_summary_by_serials(serials: List[str], access_token: AccessToken, settings: Settings):
    max_serials = 75
    req_url = 'https://apix.cisco.com/sn2info/v2/coverage/summary/serial_numbers/'
    info = get_info_by_serials(serials, access_token, req_url, max_serials, settings)
    return info


def get_psirt_by_ios_version(psirtios: List[str], access_token: AccessToken, settings: Settings):
    info = []

    headers = {'accept': 'application/json', 'Authorization': 'Bearer ' + access_token.token}
    req_url = 'https://apix.cisco.com/security/advisories/ios?version='

    # requests.packages.urllib3.disable_warnings()
    if list(psirtios) is not []:
        for ios_version in psirtios:
            log_message('request ios_version: %s, time: %s' % (ios_version, time.asctime(time.localtime(time.time()))))
            psirt_response = requests.get(req_url + ios_version, headers=headers, proxies=settings.proxies)
            if psirt_response.ok:
                log_message(
                    f'ok. ios_version: {ios_version}, '
                    f'time: {time.asctime(time.localtime(time.time()))}, '
                    f'len: {len(str(psirt_response))}'
                )
                # makes json.loads() mutch more faster (from 4 min. down to 1 sec for about 2MB)
                psirt_response.encoding = 'UTF-8'
                response = (json.loads(psirt_response.text))
                log_message(
                    f'response loaded: ios_version: {ios_version}, '
                    f'time: {time.asctime(time.localtime(time.time()))}, '
                    f'len: {len(str(response))}'
                )
                info.append({'version': ios_version, 'advisories': response.get('advisories', 'notfound')})
                log_message('ciscoapi:psirt-ios-found: %s' % info)
            else:
                log_message(
                    f'notfound. ios_version: {ios_version}, '
                    f'time: {time.asctime(time.localtime(time.time()))}, '
                    f'len: {len(str(psirt_response))}'
                )
                info.append({'version': ios_version, 'advisories': 'notfound'})
                log_message('ciscoapi:psirt-ios-notfound: %s' % info)
    return info


def get_psirt_by_iosxe_version(psirtios: List[str], access_token: AccessToken, settings: Settings):
    info = []

    headers = {'accept': 'application/json', 'Authorization': 'Bearer ' + access_token.token}
    req_url = 'https://apix.cisco.com/security/advisories/iosxe?version='

    # requests.packages.urllib3.disable_warnings()
    if list(psirtios) is not []:
        for ios_version in psirtios:
            psirt_response = requests.get(req_url + ios_version, headers=headers, proxies=settings.proxies)
            if psirt_response.ok:
                psirt_response.encoding = 'UTF-8'
                response = (json.loads(psirt_response.text))
                info.append({'version': ios_version, 'advisories': response.get('advisories', 'notfound')})
                log_message(f'ciscoapi:psirt-iosxe-found: {info}')
            else:
                info.append({'version': ios_version, 'advisories': 'notfound'})
                log_message(f'ciscoapi:psirt-iosxe-notfound: {info}')
    return info


def get_psirt_by_product_family(families: List[str], access_token: AccessToken, settings: Settings):
    info = []

    headers = {'accept': 'application/json', 'Authorization': 'Bearer ' + access_token.token}
    req_url = 'https://apix.cisco.com/security/advisories/cvrf/product?product='

    # requests.packages.urllib3.disable_warnings()
    if list(families) is not []:
        for family in families:
            psirt_response = requests.get(req_url + family, headers=headers, proxies=settings.proxies)
            if psirt_response.ok:
                psirt_response.encoding = 'UTF-8'
                response = (json.loads(psirt_response.text))
                info.append({'family': family, 'advisories': response.get('advisories', 'notfound')})
                log_message('ciscoapi:psirt-family-found: %s' % info)
            else:
                info.append({'family': family, 'advisories': 'notfound'})
                log_message('ciscoapi:psirt-family-notfound: %s' % info)
    return info


def get_suggested_release_by_pid(pids: List[str], access_token: AccessToken, settings: Settings):
    max_pid_length = 240
    max_pids = 10
    optimisedpids = []
    pidstr = ''
    count = 1
    info = []

    # split list of PIDs in chunks of max 240 bytes length
    for pid in pids:
        if (len(pidstr) + len(pid)) < max_pid_length:
            pidstr += pid + ","
            count += 1
        if (count == max_pids) or ((len(pidstr) + len(pid)) >= max_pid_length):
            optimisedpids.append(pidstr[:-1])
            pidstr = ''
            count = 1
    optimisedpids.append(pidstr[:-1])

    headers = {'accept': 'application/json', 'Authorization': 'Bearer ' + access_token.token}
    req_url = 'https://apix.cisco.com/software/suggestion/v2/suggestions/releases/productIds/'

    # requests.packages.urllib3.disable_warnings()
    if optimisedpids is not []:
        for productids in optimisedpids:
            suggestion_response = requests.get(req_url + productids, headers=headers, proxies=settings.proxies)
            log_message('cisco-api:get_sugessted_release_by_pid:response:response.text %s' % suggestion_response.text)
            if suggestion_response.ok:
                suggestion_response.encoding = 'UTF-8'
                response = (json.loads(suggestion_response.text))
                log_message('cisco-api:get_suggested_release_by_pid:response: %s' % response)
                info.append(response.get('productList', 'notfound'))
                log_message('ciscoapi:info-found: %s' % info)

    return info


# get_clean_sn_for_bug_api('ASA5510', '9.1(7)15,8.4(7)30,9.1(6)1')
# return {'9.1(7)15': '9.1(7.15)', '9.1(6)1': '9.1(6.1)', '8.4(7)30': '8.4(7.30)'}
def get_clean_sn_for_bug_api(pid, software_releases):
    if software_releases == '' or pid == '':
        return {}
    software_releases = software_releases.split(',')
    clean_sns = {}

    if pid.startswith('ASA'):
        # change ASA version from 9.1(2)10 to 9.1(2.10)
        for software_release in software_releases:
            clean_sn = software_release
            if clean_sn[-1] != ')':
                clean_sn = clean_sn.split(')')
                if len(clean_sn) == 2:
                    clean_sn = '%s.%s)' % (clean_sn[0], clean_sn[1])
            clean_sns.update({software_release: clean_sn})

    elif pid.startswith('AIR'):
        # change WLC version from 8.5.120.0 to 8.5(120.0)
        for software_release in software_releases:
            clean_sn = software_release
            if clean_sn[-1] != ')':
                if len(clean_sn.split('.')) == 4:
                    clean_sn = clean_sn.split('.')
                    clean_sn = '%s.%s(%s.%s)' % (clean_sn[0], clean_sn[1], clean_sn[2], clean_sn[3])
            clean_sns.update({software_release: clean_sn})

    else:
        for software_release in software_releases:
            clean_sn = software_release
            if clean_sn.startswith('16.0'):
                # remove leading zeros 16.09.01 to 16.9.1
                clean_sn = clean_sn.split('.')
                for x in range(0, len(clean_sn)):
                    clean_sn[x] = clean_sn[x].lstrip('0')
                    if clean_sn[x] == '':
                        clean_sn = '0'
                clean_sn = '.'.join(clean_sn)
                # remove trailing A-Za-z
                while clean_sn[-1].upper() in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
                    clean_sn = clean_sn[:-1]
            elif clean_sn.startswith('03.0'):
                # convert version from 03.03.04SE to 3.3(4)SE
                clean_sn = clean_sn.split('.')
                for x in range(0, len(clean_sn)):
                    clean_sn[x] = clean_sn[x].lstrip('0')
                    if clean_sn[x] == '':
                        clean_sn = '0'
                digits = ''
                for x in clean_sn[2]:
                    if x.isdigit():
                        digits += x
                clean_sn[2] = clean_sn[2].replace(digits, '')
                clean_sn = '%s.%s(%s)%s' % (clean_sn[0], clean_sn[1], digits, clean_sn[2])
            clean_sns.update({software_release: clean_sn})

    return clean_sns


def get_bug_by_pid_and_release(pid, release, access_token: AccessToken, reqoptions, settings: Settings):
    info = []

    headers = {'accept': 'application/json', 'Authorization': 'Bearer ' + access_token.token}
    req_url = 'https://apix.cisco.com/bug/v2.0/bugs/products/product_id/'

    if release != '':
        software_releases = get_clean_sn_for_bug_api(pid, release)
        clean_sn = ','.join(software_releases.values())
        missing = {}
        bug_response = requests.get(
            url=req_url + f'{pid}/software_releases/{clean_sn}{reqoptions}',
            headers=headers,
            proxies=settings.proxies
        )
        if bug_response.ok:
            bug_response.encoding = 'UTF-8'
            response = (json.loads(bug_response.text))
            bug_list = response.get('bugs')
            pagination_record = response.get('pagination_response_record')
            last_page = int(pagination_record.get('last_index'))
            total_records = int(pagination_record.get('total_records'))
            log_message(message=f'PID: {pid}, Version: {clean_sn}, Total records: {total_records}, Pages: {last_page}',
                        level='INFO')
            if last_page > 1:
                if reqoptions != '':
                    reqoptions = reqoptions + '&page_index='
                else:
                    reqoptions = '?page_index='
                for page in range(2, last_page + 1):
                    # time.sleep(2)
                    page_options = reqoptions + '%s' % page
                    bug_response = requests.get(req_url + f'{pid}/software_releases/{clean_sn}{page_options}',
                                                headers=headers, proxies=settings.proxies)
                    if bug_response.ok:
                        response = (json.loads(bug_response.text))
                        bug_list += response.get('bugs')
                    else:
                        bug_response.encoding = 'UTF-8'
                        status_code = bug_response.status_code
                        reason = bug_response.reason
                        url = bug_response.url
                        text = bug_response.text
                        log_message(message=f'ciscoapi error: {status_code}, {reason}, Page: {page}, LastPage: '
                                            f'{last_page}, URL: \'{url}\'. Text: \'{text}\'',
                                    level='WARNING')
                        missing.update({page: {'status_code': status_code,
                                               'reason': reason,
                                               'url': url,
                                               'text': text}})
            info.append({'pid': pid,
                         'software_releases': software_releases,
                         'bugs': bug_list,
                         'total_records': total_records,
                         'missing': missing})
            log_message('ciscoapi:bug-found: %s' % info)
        else:
            bug_response.encoding = 'UTF-8'
            status_code = bug_response.status_code
            reason = bug_response.reason
            url = bug_response.url
            text = bug_response.text
            page = 'ALL'
            log_message(f'ciscoapi error: {status_code}, {reason}, URL: \'{url}\'. Text: \'{text}\'', level='WARNING')
            missing.update({page: {'status_code': status_code,
                                   'reason': reason,
                                   'url': url,
                                   'text': text}})
            info.append({'pid': pid,
                         'software_releases': software_releases,
                         'status_code': status_code,
                         'reason': reason,
                         'missing': missing})
    return info


def get_product_mdf_information_by_pid(pids, access_token: AccessToken, settings: Settings):
    max_pid_length = 40
    max_pids = 5
    optimisedpids = []
    pidstr = ''
    count = 1
    info = []

    # split list of PIDs in chunks of max 240 bytes length
    for pid in pids:
        if (len(pidstr) + len(pid)) < max_pid_length:
            pidstr += pid + ","
            count += 1
        if (count == max_pids) or ((len(pidstr) + len(pid)) >= max_pid_length):
            optimisedpids.append(pidstr[:-1])
            pidstr = ''
            count = 1
    optimisedpids.append(pidstr[:-1])

    headers = {'accept': 'application/json', 'Authorization': 'Bearer ' + access_token.token}
    req_url = 'https://apix.cisco.com/product/v1/information/product_ids_mdf/'
#    req_url = 'https://apix.cisco.com/product/v1/information/product_ids/'

    # requests.packages.urllib3.disable_warnings()
    if optimisedpids is not []:
        for productids in optimisedpids:
            productinfo_response = requests.get(req_url + productids, headers=headers, proxies=settings.proxies)
            log_message(
                f'cisco-api:get_product_mdf_information_by_pid:response:response.text {productinfo_response.text}'
            )
            if productinfo_response.ok:
                productinfo_response.encoding = 'UTF-8'
                response = (json.loads(productinfo_response.text))
                log_message('cisco-api:get_product_mdf_information_by_pid:response: %s' % response)
                # info.append(response)
                info.extend(response.get('product_list', 'notfound'))
                log_message('ciscoapi:info-found: %s' % info)

    return info
